;; Pipeline description for Motorola PowerPC 8540 processor.
;;   Copyright (C) 2003 Free Software Foundation, Inc.
;;
;; This file is part of GCC.

;; GCC is free software; you can redistribute it and/or modify it
;; under the terms of the GNU General Public License as published
;; by the Free Software Foundation; either version 2, or (at your
;; option) any later version.

;; GCC is distributed in the hope that it will be useful, but WITHOUT
;; ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
;; or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
;; License for more details.

;; You should have received a copy of the GNU General Public License
;; along with GCC; see the file COPYING.  If not, write to the
;; Free Software Foundation, 59 Temple Place - Suite 330, Boston,
;; MA 02111-1307, USA.

(define_automaton "ppcelf8_most,ppcelf8_long,ppcelf8_retire")
(define_cpu_unit "ppcelf8_decode_0,ppcelf8_decode_1" "ppcelf8_most")

;; We don't simulate general issue queue (GIC).  If we have SU insn
;; and then SU1 insn, they can not be issued on the same cycle
;; (although SU1 insn and then SU insn can be issued) because the SU
;; insn will go to SU1 from GIC0 entry.  Fortunately, the first cycle
;; multipass insn scheduling will find the situation and issue the SU1
;; insn and then the SU insn.
(define_cpu_unit "ppcelf8_issue_0,ppcelf8_issue_1"   "ppcelf8_most")

;; We could describe completion buffers slots in combination with the
;; retirement units and the order of completion but the result
;; automaton would behave in the same way because we can not describe
;; real latency time with taking in order completion into account.
;; Actually we could define the real latency time by querying reserved
;; automaton units but the current scheduler uses latency time before
;; issuing insns and making any reservations.
;;
;; So our description is aimed to achieve a insn schedule in which the
;; insns would not wait in the completion buffer.
(define_cpu_unit "ppcelf8_retire_0,ppcelf8_retire_1" "ppcelf8_retire")

;; Branch unit:
(define_cpu_unit "ppcelf8_bu" "ppcelf8_most")

;; SU:
(define_cpu_unit "ppcelf8_su0_stage0,ppcelf8_su1_stage0" "ppcelf8_most")

;; We could describe here MU subunits for float multiply, float add
;; etc.  But the result automaton would behave the same way as the
;; described one pipeline below because MU can start only one insn
;; per cycle.  Actually we could simplify the automaton more not
;; describing stages 1-3, the result automata would be the same.
(define_cpu_unit "ppcelf8_mu_stage0,ppcelf8_mu_stage1" "ppcelf8_most")
(define_cpu_unit "ppcelf8_mu_stage2,ppcelf8_mu_stage3" "ppcelf8_most")
(define_cpu_unit "ppcelf8_mu_stage4" "ppcelf8_most")

;; The following unit is used to describe non-pipelined division.
(define_cpu_unit "ppcelf8_mu_div" "ppcelf8_long")

;; Here we simplified LSU unit description not describing the stages.
(define_cpu_unit "ppcelf8_lsu" "ppcelf8_most")

;; The following units are used to make automata deterministic
(define_cpu_unit "present_ppcelf8_decode_0" "ppcelf8_most")
(define_cpu_unit "present_ppcelf8_issue_0" "ppcelf8_most")
(define_cpu_unit "present_ppcelf8_retire_0" "ppcelf8_retire")
(define_cpu_unit "present_ppcelf8_su0_stage0" "ppcelf8_most")

;; The following sets to make automata deterministic when option ndfa is used.
(presence_set "present_ppcelf8_decode_0" "ppcelf8_decode_0")
(presence_set "present_ppcelf8_issue_0" "ppcelf8_issue_0")
(presence_set "present_ppcelf8_retire_0" "ppcelf8_retire_0")
(presence_set "present_ppcelf8_su0_stage0" "ppcelf8_su0_stage0")

;; Some useful abbreviations.
(define_reservation "ppcelf8_decode"
    "ppcelf8_decode_0|ppcelf8_decode_1+present_ppcelf8_decode_0")
(define_reservation "ppcelf8_issue"
    "ppcelf8_issue_0|ppcelf8_issue_1+present_ppcelf8_issue_0")
(define_reservation "ppcelf8_retire"
   "ppcelf8_retire_0|ppcelf8_retire_1+present_ppcelf8_retire_0")
(define_reservation "ppcelf8_su_stage0"
   "ppcelf8_su0_stage0|ppcelf8_su1_stage0+present_ppcelf8_su0_stage0")

;; Simple SU insns
(define_insn_reservation "ppcelf8_su" 1
  (and (eq_attr "type" "integer,insert_word,cmp,compare,delayed_compare,fast_compare")
       (eq_attr "cpu" "ppcelf8"))
  "ppcelf8_decode,ppcelf8_issue+ppcelf8_su_stage0+ppcelf8_retire")

;; Branch.  Actually this latency time is not used by the scheduler.
(define_insn_reservation "ppcelf8_branch" 1
  (and (eq_attr "type" "jmpreg,branch")
       (eq_attr "cpu" "ppcelf8"))
  "ppcelf8_decode,ppcelf8_bu,ppcelf8_retire")

;; Multiply
(define_insn_reservation "ppcelf8_multiply" 4
  (and (eq_attr "type" "imul,imul2,imul3,imul_compare")
       (eq_attr "cpu" "ppcelf8"))
  "ppcelf8_decode,ppcelf8_issue+ppcelf8_mu_stage0,ppcelf8_mu_stage1,\
   ppcelf8_mu_stage2,ppcelf8_mu_stage3+ppcelf8_retire")

;; Divide.  We use the average latency time here.  We omit reserving a
;; retire unit because of the result automata will be huge.  We ignore
;; reservation of miu_stage3 here because we use the average latency
;; time.
(define_insn_reservation "ppcelf8_divide" 14
  (and (eq_attr "type" "idiv")
       (eq_attr "cpu" "ppcelf8"))
  "ppcelf8_decode,ppcelf8_issue+ppcelf8_mu_stage0+ppcelf8_mu_div,\
   ppcelf8_mu_div*13")

;; CR logical
(define_insn_reservation "ppcelf8_cr_logical" 1
  (and (eq_attr "type" "cr_logical,delayed_cr")
       (eq_attr "cpu" "ppcelf8"))
  "ppcelf8_decode,ppcelf8_bu,ppcelf8_retire")

;; Mfcr
(define_insn_reservation "ppcelf8_mfcr" 1
  (and (eq_attr "type" "mfcr")
       (eq_attr "cpu" "ppcelf8"))
  "ppcelf8_decode,ppcelf8_issue+ppcelf8_su1_stage0+ppcelf8_retire")

;; Mtcrf
(define_insn_reservation "ppcelf8_mtcrf" 1
  (and (eq_attr "type" "mtcr")
       (eq_attr "cpu" "ppcelf8"))
  "ppcelf8_decode,ppcelf8_issue+ppcelf8_su1_stage0+ppcelf8_retire")

;; Mtjmpr
(define_insn_reservation "ppcelf8_mtjmpr" 1
  (and (eq_attr "type" "mtjmpr,mfjmpr")
       (eq_attr "cpu" "ppcelf8"))
  "ppcelf8_decode,ppcelf8_issue+ppcelf8_su_stage0+ppcelf8_retire")

;; Loads
(define_insn_reservation "ppcelf8_load" 3
  (and (eq_attr "type" "load,load_ext,load_ext_u,load_ext_ux,load_ux,load_u")
       (eq_attr "cpu" "ppcelf8"))
  "ppcelf8_decode,ppcelf8_issue+ppcelf8_lsu,nothing,ppcelf8_retire")

;; Stores.
(define_insn_reservation "ppcelf8_store" 3
  (and (eq_attr "type" "store,store_ux,store_u")
       (eq_attr "cpu" "ppcelf8"))
  "ppcelf8_decode,ppcelf8_issue+ppcelf8_lsu,nothing,ppcelf8_retire")

;; Simple FP
(define_insn_reservation "ppcelf8_simple_float" 1
  (and (eq_attr "type" "fpsimple")
       (eq_attr "cpu" "ppcelf8"))
  "ppcelf8_decode,ppcelf8_issue+ppcelf8_su_stage0+ppcelf8_retire")

;; FP
(define_insn_reservation "ppcelf8_float" 4
  (and (eq_attr "type" "fp")
       (eq_attr "cpu" "ppcelf8"))
  "ppcelf8_decode,ppcelf8_issue+ppcelf8_mu_stage0,ppcelf8_mu_stage1,\
   ppcelf8_mu_stage2,ppcelf8_mu_stage3+ppcelf8_retire")

(define_insn_reservation "ppcelf8_dp_float" 5
  (and (eq_attr "type" "fpdouble")
       (eq_attr "cpu" "ppcelf8"))
  "ppcelf8_decode,ppcelf8_issue+ppcelf8_mu_stage0,ppcelf8_mu_stage1,\
   ppcelf8_mu_stage2,ppcelf8_mu_stage3,ppcelf8_mu_stage4+ppcelf8_retire")

;; float divides.  We omit reserving a retire unit and miu_stage3
;; because of the result automata will be huge.
(define_insn_reservation "ppcelf8_float_vector_divide" 29
  (and (eq_attr "type" "vecfdiv")
       (eq_attr "cpu" "ppcelf8"))
  "ppcelf8_decode,ppcelf8_issue+ppcelf8_mu_stage0+ppcelf8_mu_div,\
   ppcelf8_mu_div*28")

;; DP FP Divides are not pipelined
(define_insn_reservation "ppcelf8_double_float__divide" 32
  (and (eq_attr "type" "fpdoublediv")
       (eq_attr "cpu" "ppcelf8"))
  "ppcelf8_decode,ppcelf8_issue+ppcelf8_mu_stage0+ppcelf8_mu_div,\
   ppcelf8_mu_div*31")

;; Brinc
(define_insn_reservation "ppcelf8_brinc" 1
  (and (eq_attr "type" "brinc")
       (eq_attr "cpu" "ppcelf8"))
  "ppcelf8_decode,ppcelf8_issue+ppcelf8_su_stage0+ppcelf8_retire")

;; Simple vector
(define_insn_reservation "ppcelf8_simple_vector" 1
  (and (eq_attr "type" "vecsimple")
       (eq_attr "cpu" "ppcelf8"))
  "ppcelf8_decode,ppcelf8_issue+ppcelf8_su1_stage0+ppcelf8_retire")

;; Simple vector compare
(define_insn_reservation "ppcelf8_simple_vector_compare" 1
  (and (eq_attr "type" "veccmpsimple")
       (eq_attr "cpu" "ppcelf8"))
  "ppcelf8_decode,ppcelf8_issue+ppcelf8_su_stage0+ppcelf8_retire")

;; Vector compare
(define_insn_reservation "ppcelf8_vector_compare" 1
  (and (eq_attr "type" "veccmp")
       (eq_attr "cpu" "ppcelf8"))
  "ppcelf8_decode,ppcelf8_issue+ppcelf8_su1_stage0+ppcelf8_retire")

;; evsplatfi evsplati
(define_insn_reservation "ppcelf8_vector_perm" 1
  (and (eq_attr "type" "vecperm")
       (eq_attr "cpu" "ppcelf8"))
  "ppcelf8_decode,ppcelf8_issue+ppcelf8_su1_stage0+ppcelf8_retire")

;; Vector float
(define_insn_reservation "ppcelf8_float_vector" 4
  (and (eq_attr "type" "vecfloat")
       (eq_attr "cpu" "ppcelf8"))
  "ppcelf8_decode,ppcelf8_issue+ppcelf8_mu_stage0,ppcelf8_mu_stage1,\
   ppcelf8_mu_stage2,ppcelf8_mu_stage3+ppcelf8_retire")

;; Vector divides: Use the average.  We omit reserving a retire unit
;; because of the result automata will be huge.  We ignore reservation
;; of miu_stage3 here because we use the average latency time.
(define_insn_reservation "ppcelf8_vector_divide" 14
  (and (eq_attr "type" "vecdiv")
       (eq_attr "cpu" "ppcelf8"))
  "ppcelf8_decode,ppcelf8_issue+ppcelf8_mu_stage0+ppcelf8_mu_div,\
   ppcelf8_mu_div*13")

;; Complex vector.
(define_insn_reservation "ppcelf8_complex_vector" 4
  (and (eq_attr "type" "veccomplex")
       (eq_attr "cpu" "ppcelf8"))
  "ppcelf8_decode,ppcelf8_issue+ppcelf8_mu_stage0,ppcelf8_mu_stage1,\
   ppcelf8_mu_stage2,ppcelf8_mu_stage3+ppcelf8_retire")

;; Vector load
(define_insn_reservation "ppcelf8_vector_load" 3
  (and (eq_attr "type" "vecload")
       (eq_attr "cpu" "ppcelf8"))
  "ppcelf8_decode,ppcelf8_issue+ppcelf8_lsu,nothing,ppcelf8_retire")

;; Vector store
(define_insn_reservation "ppcelf8_vector_store" 3
  (and (eq_attr "type" "vecstore")
       (eq_attr "cpu" "ppcelf8"))
  "ppcelf8_decode,ppcelf8_issue+ppcelf8_lsu,nothing,ppcelf8_retire")
